package WWW::OrangeHRM::Client;
use strict;
use warnings;
use utf8;
use version 0.77; our $VERSION = qv("v0.5.0");

use WWW::Mechanize;
use HTML::TreeBuilder::LibXML;
use DateTime;
use DateTime::Duration;

use Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(debug_http dump_state fatal_error log_in log_out time_sheet
    time_sheet_change time_sheet_date time_sheet_year_month time_sheet_show
    normalize_duration normalize_time time_sheet_set time_sheet_submit
    year_month2date);

=encoding utf8

=head1 NAME

WWW::OrangeHRM::Client - Client for OrangeHRM

=head1 DESCRIPTION

This module implements client for
L<OrangeHRM|http://en.wikipedia.org/wiki/OrangeHRM> information system. It has
been developed against Red Hat instance, so I cannot guarantee it works with
other instances. Author and this code have no business or personal relation to
OrangeHRM Inc.

=head1 SYNOPSIS

    use WWW::OrangeHRM::Client;
    my %configuration = (
        url => 'https://redhat.orangehrm.com/',
        username => 'foo',
        password => 'bar'
    );
    my $debug = 0;
    my $automaton = WWW::Mechanize->new();
    
    # Log-in
    if (!WWW::OrangeHRM::Client::log_in($automaton, \%configuration)) {
        WWW::OrangeHRM::Client::fatal_error($automaton, $debug,
            'Could not log in!');
    }
    print "Logged in.\n";
    
    if (!WWW::OrangeHRM::Client::time_sheet($automaton)) {
        WWW::OrangeHRM::Client::fatal_error($automaton, $debug,
            'Could not get time sheet!');
    }
    
    WWW::OrangeHRM::Client::time_sheet_show($automaton);
    
    # Log-out
    if (!WWW::OrangeHRM::Client::log_out($automaton)) {
        WWW::OrangeHRM::Client::fatal_error($automaton, $debug,
            'Could not log out!');
    }
    print "Logged out.\n";

=cut

my $last_request;

BEGIN {
    $last_request = undef;
}


=head1 API

$automaton is C<WWW::Mechanize> object.

=head2 %configuration

    my %configuration = (
        url => 'https://redhat.orangehrm.com/',
        username => 'foo@redhat.com',
        password => undef
    );

=cut 

=head2 debug_http($automaton)

Call this after creating $automaton if you want to see HTTP request in
dump_state() or fatal_errro() dump.

=cut

sub debug_http {
    my $automaton = shift;
    $automaton->add_handler("request_send", sub {
            $last_request = shift->dump(maxlength => 0);
            return;
        });
    1;
}


=head2 dump_state($automaton)

Print details about last request and response.

=cut

sub dump_state {
    my $automaton = shift;

    if (defined $last_request) {
        print "===BEGIN HTTP REQUEST===\n";
        print $last_request;
        print "===END HTTP REQUEST===\n";
    }

    print 'Current URI: ', $automaton->uri, "\n";

    print "===BEGIN RESPONSE HEADERS===\n";
    $automaton->dump_headers();
    print "===END RESPONSE HEADERS===\n";

    print "===BEGIN RESPONSE BODY===\n";
    print $automaton->response->decoded_content, "\n";
    print "===END RESPONSE BODY===\n";
    
    print "===BEGIN TEXT DUMP===\n";
    $automaton->dump_text();
    print "===END TEXT DUMP===\n";
    
    print "===BEGIN LINK DUMP===\n";
    $automaton->dump_links();
    print "===END LINK DUMP===\n";
    
    print "===BEGIN FORM DUMP===\n";
    $automaton->dump_forms();
    print "===END FORM DUMP===\n";

    return 1;
}


=head2 fatal_error($automaton, $debug, $message)

Call this function, if you encounter fatal error.

$message will be printed. Automaton state will be dumped if $debug is true.

This function does not return, it exits program with an error exit code.

=cut

sub fatal_error {
    my ($automaton, $debug, $message) = @_;

    print $message, "\n";
    if ($debug) {
        dump_state($automaton);
        print $message, "\n";
    }
    
    exit 1;
}


=head2 get_text_status($automaton)

Return text in *[@id='messageBalloonContainer']/div element of current HTML
document or undef.

The @id specified element exists always. It always contains white space
text node. But the error message is stored in I<div> sub-element children and
the sub-element exists when reporting error.

=cut

sub get_text_status {
    my $automaton = shift;
    my $document = $automaton->response->decoded_content;
    my $parser = HTML::TreeBuilder::LibXML->new()->parse_content($document)
        or return undef;
    my $nodes = $parser->findnodes(q{//*[@id="messageBalloonContainer"]/div});
    # $parser->findvalue() returns q{} if XPath does not match. There is no
    # way how to distinguish from element-only subtree. Therefore use size().
    my $text;
    if ($nodes->size > 0) {
        $text = $nodes->to_literal;
    }
    return $text;
}


=head2 log_in($automaton, \%configuration)

Log into system and user as defined in %configuration.

Return false in case of failure, true in case of success.

=cut

sub log_in {
    my ($automaton, $configuration) = @_;

    print "Getting log-in page for <", $configuration->{url}, ">\n";
    $automaton->get($configuration->{url});

    print "Logging in as `", $configuration->{username}, "'\n";
    $automaton->submit_form(
        form_name => 'frmLogin',
        fields  => {
            'login[txtUserName]' => $configuration->{username},
            'login[txtUserPassword]' => $configuration->{password}
        },
    );
    if ($automaton->uri !~ m{/leave/redHatViewLeaveList\z}) {
        return 0;
    }
    return 1;
}


=head2 log_out($automaton)

Log out the system.

Return false in case of failure, true in case of success.

=cut

sub log_out {
    my $automaton = shift;

    print "Logging out\n";
    $automaton->get('/index.php/auth/logOut');

    if ($automaton->uri !~ m{/index.php/\z}) {
        return 0;
    }

    return 1;
}


=head2 time_sheet_year_month($automaton)

Return time identification of current time sheet as (year, month) numbers pair.
E.g. C<(2012, 12)>.
Return undef in case of error or partial list, e.g. C<(2012, undef)>.

=cut

sub time_sheet_year_month {
    my $automaton = shift;

    my $form = $automaton->form_id('editTimesheet') or return (undef, undef);

    return (
        $form->value('editTimesheet[yearChoices]'),
        $form->value('editTimesheet[monthChoices]')
    );
}


=head2 year_month2date($year, $month)

Return year and month formated to string. E.g. C<2012-12>.
Return undef in case of error or partial string, e.g. C<< 2012-<unknown month>
>>.

=cut

sub year_month2date {
    my ($year, $month) = @_;
    if (!defined $year and !defined $month) { return undef; }
    return
        ($year // '<unknown year>') . '-' .  ($month // '<uknown month>');
}


=head2 time_sheet_date($automaton)

Return time identification of current time sheet as string. E.g. C<2012-12>.
Return undef in case of error or partial string, e.g. C<< 2012-<unknown month>
>>.

=cut

sub time_sheet_date {
    return year_month2date(time_sheet_year_month(shift));
}


=head2 time_sheet($automaton)

Retrieve current time sheet.

Return false in case of failure, true in case of success.

=cut

sub time_sheet {
    my $automaton = shift;
    my $path = '/index.php/time/editTimesheet/show/mine';

    print "Getting time sheet\n";
    $automaton->get($path);

    if ($automaton->uri !~ m{$path\z}) {
        return 0;
    }

    print "Time sheet retrieved successfully.\n";
    return 1;
}


=head2 time_sheet_change($automaton, $year, $month)

Retrieve time sheet for month specified by $year and $month. Month counts from
1, year is in Gregorian calendar.

Return false in case of failure, true in case of success.

=cut

sub time_sheet_change {
    my ($automaton, $year, $month) = @_;
    my $path = '/index.php/time/editTimesheet(?:/show/mine)?';

    if (!defined $year || !defined $month) {
        return 0;
    }
    if ($automaton->uri !~ m{$path\z}) {
        return 0;
    }

    print "Changing time sheet month to $year-$month\n";
    my $form = $automaton->form_id('searchTimesheet') or return 0;
    $automaton->set_fields(
        'editTimesheet[yearChoices]' => $year,
        'editTimesheet[monthChoices]' => $month
    );
    my $submit_button = $form->find_input('#btnSearch') or return 0;
    $automaton->click_button(input => $submit_button);

    if ($automaton->uri !~ m{$path\z}) {
        return 0;
    }

    if (defined (my $message = get_text_status($automaton))) {
        print "Changing time sheet failed: $message\n";
        return 0;
    }
    my $date = time_sheet_date($automaton);
    if (!defined $date or !($date eq "$year-$month")) {
        return 0;
    }

    print "Time sheet changed successfully.\n";
    return 1;
}


=head2 time_sheet_show($automaton)

Show current time sheet.

Return false in case of failure, true in case of success.

=cut

sub time_sheet_show {
    my $automaton = shift;
    my $path = '/index.php/time/editTimesheet(?:/show/mine)?';

    if ($automaton->uri !~ m{$path\z}) {
        return 0;
    }

    #dump_state($automaton);

    my $document = $automaton->response->decoded_content;
    my $parser = HTML::TreeBuilder::LibXML->new()->parse_content($document)
        or return 0;
    my $form = $automaton->form_id('editTimesheet') or return 0;

    my ($year, $month) = time_sheet_year_month($automaton);
    my $date = year_month2date($year, $month);
    if (defined $date) {
        printf "Time sheet for %s:\n", $date;
    }
    my $day;
    if (defined $year and defined $month) {
        $day = DateTime->new(year => $year, month => $month);
    }

    # Search table body only. It's 2 times faster than whole document;
    my $time_sheet_body = $parser->findnodes(
        q{//table[@id="timesheetTable"]/tbody}
    );
    if ($time_sheet_body->size != 1) { return 0; }
    $time_sheet_body = $time_sheet_body->get_node(0);

    print "Day\tDayType\tWhere\tFrom\tTo\tBreak\tDoctor\tTotal\tComment\n";
    for (map {$_->name} ($form->inputs)) {
        next if ! defined;
        if (/\Aoriginal_record_ids\[([0,1,2,3,4,5,6,7,8,9]+)\]\z/) {
            my $index = $1;
            my $day_number = $index + 1;

            if (defined $day) {
                # XXX Do not trust $day_number from input.
                eval { $day->set(day => $day_number) };
                if ($@) {
                    print $day_number;
                } else  {
                    print $day->strftime('%d %a');
                }
            } else {
                print $day_number;
            }

            my $class = $time_sheet_body->findvalue(
                q{tr[td/input/@id="record_id_}. $index . q{"]/@class}
            ) // '';
            if ($class eq 'days_off') {
                $class = 'weekend';
            } elsif ($class eq 'full_day_holiday') {
                $class = 'holiday';
            } elsif ($class eq 'leave_day') {
                # ??? What is it?
                $class = 'sickday';
            } elsif ($class eq 'leave_date') {
                # This is both sickday and leave.
                # TODO: Discriminate on table[@class=~'leave_info_table']
                $class = 'leave'
            } elsif ($class eq 'half_day_leave') {
                $class = 'leave/2'
            }
            print "\t$class";

            if ($form->value("classification_id\[$index\]") == 2) {
                print "\ttrip";
            } else {
                print "\twork";
            }

            for (qw(in_time out_time break_duration medical_checkup_duration
                row_total_duration comment)) {
                print "\t", $form->value("$_\[$index\]") // '';
            }

            print "\n";
        }
    }


    printf "Total leave: %s\n",
        normalize_duration($parser->findvalue(q{//*[@id="totalLeave"]/@value}))
        // '';
    printf "Total time entered: %s\n",
        normalize_duration($parser->findvalue(
            q{//th[text()="Total Time Entered"]/following-sibling::td[1]}
        )) // '';
    printf "Total time: %s\n",
        normalize_duration($parser->findvalue(q{//*[@id="total"]/@value}))
        // '';
    printf "Total time expected: %s\n",
        normalize_duration($parser->findvalue(q{//*[@id="expected"]/@value}))
        // '';

    my $status =
        $parser->findvalue(q{//*[@id="editSection"]/div[1]/div[@class="heading"]});
    if (defined $status) {
        $status =~ s/\R/ /s;
        $status =~ s/.*\s+Status:\s*(.*\w)\s*/$1/;
        printf "Status: %s\n", $status;
    }
    return 1;
}


=head2 normalize_time($time)

Return $time value normalized to HH:MM format.

Return undef if $time is invalid.

This function can differ from web interface.

=cut

sub normalize_time {
    $_ = shift;
    if (!defined) {
        return undef;
    }

    my ($hours, $minutes);
    if (($hours) = (m/\A([0123456789]*)\z/)) {
        $minutes = '00';
    } elsif (!(  ($hours, $minutes) =
            (m/\A([0123456789]*):([0123456789]*)\z/)
        )) {
        print STDERR "Invalid time value `$_'\n";
        return undef;
    }
    
    for ($hours, $minutes) {
        if (length($_) == 0) {
            $_ = '00';
        } elsif (length($_) == 1) {
            $_ = '0' . $_;
        }
    }

    return $hours . ':' . $minutes;
}


=head2 normalize_duration($duration, $strict)

Return $duration value normalized to HH:MM format.

If $duration is invalid, return original value if $strict is false, return
undef if $strict is true.

This function can differ from web interface.

=cut

sub normalize_duration {
    my ($_, $strict) = @_;
    if (!defined) {
        return undef;
    }

    my ($hours, $minutes);
    if (($hours) = (m/\A([0123456789]*)\z/)) {
        $minutes = '00';
    } elsif (!(  ($hours, $minutes) =
            (m/\A([0123456789]*):([0123456789]*)\z/)
        )) {
        print STDERR "Invalid duration value `$_'\n";
        if ($strict) {
            return undef;
        } else {
            return $_;
        }
    }
    
    my $duration = DateTime::Duration->new(
        hours => $hours, minutes=> $minutes);
    return sprintf('%02d:%02d', $duration->in_units(q{hours}, q{minutes}));
}


=head2 set_if_not_amend_or_defined($amend, $field, $value)

Return list ($field, $value) if $amend is false or $value is defined. Otherwise return empty list.

=cut

sub set_if_not_amend_or_defined {
    my ($amend, $field, $value) = @_;
    (!$amend or defined $value) ? ( $field => $value ) : ();
}


=head2 time_sheet_set($automaton, $day, $from, $to, $break, $doctor,
$comment, $trip, $amend)

Fill a day into time sheet as if you worked from $from to $to with $break
and $doctor hours. $day of month counts from 1. Time format is HH:MM. $comment
is free text, $trip is true for bussines trip, otherwise standard work.

If #amend is false, the day will be overwritten, i.e undefined arguments will
empty the day records. If $amend is true, only defined arguments will set,
keeping records with undefined arguments untouched.

Return false in case of failure, true in case of success.

=cut

sub time_sheet_set {
    my ($automaton, $day, $from, $to, $break, $doctor, $comment, $trip, $amend)
        = @_;
    my $index = $day - 1;
    my $classification;
    if (defined $trip or !$amend) {
        $classification = ($trip ? 2 : 1);
    }
    if (defined $doctor) {
        # Server refuses H:MM $doctor value
        $doctor = normalize_time($doctor);
    }

    print "Filling day #", $day, " into time sheet\n";

    my $form = $automaton->form_id('editTimesheet') or return 0;
    #print "DEBUG1: ", $form->dump(), "\n";
    #print "DEBUG amend: ", Data::Dumper::Dumper(set_if_not_amend_or_defined($amend,
    #        "classification_id[$index]" => $classification)), "\n";
    $automaton->set_fields(
        set_if_not_amend_or_defined($amend,
            "classification_id[$index]" => $classification),
        set_if_not_amend_or_defined($amend,
            "in_time[$index]" => $from),
        set_if_not_amend_or_defined($amend,
            "out_time[$index]" => $to),
        set_if_not_amend_or_defined($amend,
            "break_duration[$index]" => $break),
        set_if_not_amend_or_defined($amend,
            "medical_checkup_duration[$index]" => $doctor),
        set_if_not_amend_or_defined($amend,
            "comment[$index]" => $comment),
        # This is set by Firefox magically. Probably a JS voodoo.
        #"row_total_duration[$index]" not necessary, server computes it.
        "editTimesheet[txtAction]" => 'save'
    );
    # Server expects disabled values in the form, otherwise it reports failure
    # despite saving the data. Server disables some inputs on full-day
    # leave.
    for (grep {$_->disabled} ($form->inputs)) {
        $_->disabled(0);
    }
    #print "DEBUG2: ", $form->dump(), "\n";
    # WWW::Mechanize cannot find the 'Save' button by value for uknown reason.
    my $submit_button = $form->find_input('#btnSave') or return 0;
    $automaton->click_button(input => $submit_button);

    if ($automaton->uri !~ m{/index.php/time/editTimesheet\z}) {
        return 0;
    }
    my $message = get_text_status($automaton);
    if (! defined $message) {
        return 0;
    }
    if (! ($message eq 'Timesheet Successfully Saved') ) {
        print "Saving time sheet failed: $message\n";
        return 0;
    }

    print "Day saved successfully.\n";
    return 1;
}


=head2 time_sheet_submit($automaton, $comment)

Submit current time sheet for review to your manager. You can set submission
$comment.

Return false in case of failure, true in case of success.

=cut

sub time_sheet_submit {
    my ($automaton, $comment) = @_;

    my $date = time_sheet_date($automaton);
    if (defined $date) {
        print "Submitting time sheet for $date\n";
    } else {
        print "Submitting time sheet\n";
    }

    my $form = $automaton->form_id('editTimesheet') or return 0;
    $automaton->set_fields(
        "editTimesheet[submitComment]" => $comment,
        # JS voodoo:
        "editTimesheet[txtAction]" => 'submit'
    );
    # Server expects disabled values in the form, otherwise it reports failure
    # despite saving the data. Server disables some inputs on full-day
    # leave.
    for (grep {$_->disabled} ($form->inputs)) {
        $_->disabled(0);
    }
    my $submit_button = $form->find_input('#btnSubmit') or return 0;
    $automaton->click_button(input => $submit_button);

    if ($automaton->uri !~ m{/index.php/time/editTimesheet\z}) {
        return 0;
    }
    my $message = get_text_status($automaton);
    if (! defined $message) {
        return 0;
    }
    if (! ($message eq 'Timesheet Successfully Submitted') ) {
        print "Submitting time sheet failed: $message\n";
        return 0;
    }

    print "Time sheet submitted successfully.\n";
    return 1;
}

1;
__END__

=head1 COPYRIGHT

Copyright © 2012, 2013  Petr Písař <ppisar@redhat.com>.

=head1 LICENSE

This is free software.  You may redistribute copies of it under the terms of
the GNU General Public License L<http://www.gnu.org/licenses/gpl.html>.
There is NO WARRANTY, to the extent permitted by law.

=cut
